/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

/* All .h files include here 
   Why not use .c? Since template function now only work in .h*/
#include "include.h"

/* main */
int main(int argc, char** argv)
{   
    if (argc < 5 || argc > 8) 
    {
	/* Usage example: ./solver PCG CSR Jacobi ../3.0/UF\ collection/bcsstk16.mtx */
	/* Options in [] are not compulsory */
        printf("Correct Usage: ./solver <CG|PCG> <CSR|ELL> <NONE|Jacobi|SSOR> <input matrix file> [<input rhs file> [<input initial x file> [<output x file>]]]\n");
        exit(EXIT_FAILURE);
    }
    int max_iter = MAX_ITER;
    TYPE error_rel = ERROR_REL;

    char solver_name[32];
    strcpy(solver_name, argv[1]);
    if(strcmp(solver_name, "CG") && strcmp(solver_name, "PCG") && strcmp(solver_name, "PGMRES"))
    {
        printf("Solver must be CG|PCG|PGMRES \n");
        exit(EXIT_FAILURE);
    }

    char storage_name[32];
    strcpy(storage_name, argv[2]);
    if(strcmp(storage_name, "CSR") && strcmp(storage_name, "ELL"))
    {
        printf("Storage name must be CSR|ELL\n");
        exit(EXIT_FAILURE);
    }
    
    char precon_name[32];
    strcpy(precon_name, argv[3]);
    if(strcmp(precon_name, "NONE") && strcmp(precon_name, "Jacobi") && strcmp(precon_name, "SSOR"))
    {
        printf("Preconditioner name must be NONE|Jacobi|SSOR\n");
        exit(EXIT_FAILURE);
    }else if(!strcmp(precon_name, "NONE") && strcmp(solver_name, "CG"))
        {
            printf("Preconditioner must not be NONE!\n");
            exit(EXIT_FAILURE);            
        }
    
    char spm_filename[256], b_filename[256], x_ini_filename[256], x_filename[256];
    strcpy(spm_filename, argv[4]);
    spmatrix<TYPE> spm = read_spmatrix<TYPE>(spm_filename);

	/* Handle strings with some path. 
       input file "../3.0/UF\ collection/bcsstk16.mtx" and output file "bcsstk16_result.mtx" at current folder  
		*/
    char *UF_c = "UF_collection";
    char *ptr_UF_c = strstr(spm_filename, UF_c);//letters behind "UF collection"
    char *backslash = strtok(ptr_UF_c, "/");
    backslash = strtok(NULL, "/");//letters behind "/"
    char *dot = strtok(backslash, ".");//letters before "."
    char *_result = "_result.txt";
    strcat(dot, _result);
    char mat_result[32];
    stpcpy(mat_result, dot);
	/* The Output file */
    FILE *fpDat;
    if((fpDat = fopen(mat_result, "w")) == NULL)
    {
        printf("Can not open file %s\n", mat_result);
        exit(EXIT_FAILURE);
    }         
    fprintf(fpDat, "Matrix info: \nrowNum: %d colNum: %d nonzeroNum: %d\n\n", spm.nrow, spm.ncol, spm.nnzero);

    //for(int i = 0; i < spm.nrow; i++)
    //printf("%d \n", spm.nnz_per_row[i]);

	/* Two sparse matrix formats for SPMV: CSR, ELL */
    csr_format<TYPE> spm_csr;
    ell_format<TYPE> spm_ell;
    if(!strcmp(storage_name, "CSR"))
    {   
        fprintf(fpDat, "CSR format is used.\n");
        convert_matrix<TYPE>(&spm, &spm_csr);
        #if PRINT_MATRIX
        print_matrix<TYPE>(&spm_csr);
        #endif
    }
    else if(!strcmp(storage_name, "ELL"))
    {
        fprintf(fpDat, "ELL format is used.\n");
        convert_matrix<TYPE>(&spm, &spm_ell);
        #if PRINT_MATRIX
        print_matrix<TYPE>(&spm_ell);
        #endif        
    }
    
	/* Three sparse matrix formats for Preconditioner: Diag for Jacobi, CSC for SSOR 
	they should be defined here, however, they have not been allocated or initialized */
    spm_diag<TYPE> spm_dia;

    spmatrix<TYPE> spm_lower;
    spmatrix<TYPE> spm_upper;
    csc_triangle<TYPE> csc_lower;
    csc_triangle<TYPE> csc_upper;

    spm_array<TYPE> spm_arr_lower;   
    spm_array<TYPE> spm_arr_upper; 
    csc_tri_array<TYPE> csc_arr_lower; 
    csc_tri_array<TYPE> csc_arr_upper;        

    Jacobi_precon<TYPE> Jacobi;
    ssor_precon<TYPE> ssor;
    sub_ssor_precon<TYPE> sub_ssor;

    if(!strcmp(precon_name, "Jacobi"))
    {     
        fprintf(fpDat, "Jacobi preconditioner is used.\n\n");
        get_diag<TYPE>(&spm, &spm_dia);
        Jacobi_setup<TYPE>(&spm_dia, &Jacobi);
        clear_diag<TYPE>(&spm_dia);
    }
    else if(!strcmp(precon_name, "SSOR"))
    {
        fprintf(fpDat, "SSOR preconditioner is used.\n\n");
        get_lower<TYPE>(&spm, &spm_lower);
        get_lower<TYPE>(&spm, &spm_arr_lower);
        get_diag<TYPE>(&spm, &spm_dia);
        get_upper<TYPE>(&spm, &spm_upper);
        get_upper<TYPE>(&spm, &spm_arr_upper);        

        convert_lower<TYPE>(&spm_lower, &csc_lower);
        convert_lower<TYPE>(&spm_arr_lower, &csc_arr_lower);
        convert_upper<TYPE>(&spm_upper, &csc_upper);
        convert_upper<TYPE>(&spm_arr_upper, &csc_arr_upper);
        
        ssor_setup<TYPE>(&csc_lower, &spm_dia, &csc_upper, &ssor);
        ssor_setup<TYPE>(&csc_arr_lower, &spm_dia, &csc_arr_upper, &sub_ssor);

        #if PRINT_PRECON
        print_triangle<TYPE>(&spm_arr_lower);
        //print_triangle<TYPE>(&csc_lower);
        #endif

        //clear_csc_triangle<TYPE>(&csc_arr_lower);
        //clear_csc_triangle<TYPE>(&csc_arr_upper);         
        clear_triangle<TYPE>(&spm_upper);
        clear_triangle<TYPE>(&spm_arr_upper);
        clear_triangle<TYPE>(&spm_lower);
        clear_triangle<TYPE>(&spm_arr_lower); 
    }
    
    int nrow = spm.nrow;
    int ncol = spm.ncol;
    clear_spmatrix<TYPE>(&spm);

    TYPE *h_b;
    TYPE *h_x_ref;
    TYPE *h_x;
    TYPE *mh_x;
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **) &h_b, nrow * sizeof(TYPE)));
    CUDA_SAFE_CALL(cudaMallocHost((void **) &h_x_ref, ncol * sizeof(TYPE)));
    CUDA_SAFE_CALL(cudaMallocHost((void **) &h_x, ncol * sizeof(TYPE)));
    CUDA_SAFE_CALL(cudaMallocHost((void **) &mh_x, ncol * sizeof(TYPE)));
    #else
	h_b = (TYPE *)malloc(nrow * sizeof(TYPE));
	h_x_ref = (TYPE *)malloc(ncol * sizeof(TYPE));
	h_x = (TYPE *)malloc(ncol * sizeof(TYPE));
	mh_x = (TYPE *)malloc(ncol * sizeof(TYPE));
    #endif

    if (argc >= 6) 
    { 
        strcpy(b_filename, argv[5]);
        read_data(h_b, b_filename, nrow);
    }else
    {
        for(int i = 0; i < nrow; i++)
        h_b[i] = rand() / (TYPE)RAND_MAX;
    }
    if (argc >= 7) 
    { 
        strcpy(x_ini_filename, argv[6]);
        read_data(h_x_ref, x_ini_filename, ncol);
        read_data(h_x, x_ini_filename, ncol);
        read_data(mh_x, x_ini_filename, ncol);
    }else
    {
        #if SET_X_INI_ZERO
        memset(h_x_ref, 0, ncol * sizeof(TYPE));
        memset(h_x, 0, ncol * sizeof(TYPE));
        memset(mh_x, 0, ncol * sizeof(TYPE));
        #else
        for(int i = 0; i < ncol; i++)
        {
            h_x_ref[i] = rand() / (TYPE)RAND_MAX;
            h_x[i] = rand() / (TYPE)RAND_MAX;
            mh_x[i] = rand() / (TYPE)RAND_MAX;
        }
        #endif
    }
	
	/* solve Ax = b on host */
    #if EXECUTE_CPU
    int iter_cpu;
    unsigned int timer_h = 0;
    double t_h_ms = 0.0;
    TYPE res_norm_ref;
	start_timer(&timer_h);
    if(!strcmp(solver_name, "CG"))
    {   
        if(!strcmp(storage_name, "CSR"))
            res_norm_ref = cg_host<csr_format<TYPE>, TYPE>(fpDat, &spm_csr, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);
        else if(!strcmp(storage_name, "ELL"))
            res_norm_ref = cg_host<ell_format<TYPE>, TYPE>(fpDat, &spm_ell, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);
    }
    else if(!strcmp(solver_name, "PCG"))
    {
        if(!strcmp(storage_name, "CSR"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm_ref = pcg_host<csr_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_csr, &Jacobi, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);
            else if(!strcmp(precon_name, "SSOR"))
                res_norm_ref = pcg_host<csr_format<TYPE>, ssor_precon<TYPE>, TYPE>(fpDat, &spm_csr, &ssor, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);            
        }
        else if(!strcmp(storage_name, "ELL"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm_ref = pcg_host<ell_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_ell, &Jacobi, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);  
            else if(!strcmp(precon_name, "SSOR"))
                res_norm_ref = pcg_host<ell_format<TYPE>, ssor_precon<TYPE>, TYPE>(fpDat, &spm_ell, &ssor, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);              
        }
    }
    else if(!strcmp(solver_name, "PGMRES"))
    {
        if(!strcmp(storage_name, "CSR"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm_ref = pgmres_host<csr_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_csr, &Jacobi, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);
            else if(!strcmp(precon_name, "SSOR"))
                res_norm_ref = pgmres_host<csr_format<TYPE>, ssor_precon<TYPE>, TYPE>(fpDat, &spm_csr, &ssor, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);            
        }
        else if(!strcmp(storage_name, "ELL"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm_ref = pgmres_host<ell_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_ell, &Jacobi, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);  
            else if(!strcmp(precon_name, "SSOR"))
                res_norm_ref = pgmres_host<ell_format<TYPE>, ssor_precon<TYPE>, TYPE>(fpDat, &spm_ell, &ssor, h_b, h_x_ref, max_iter, error_rel, &iter_cpu);              
        }
    }    
    stop_timer(timer_h, &t_h_ms);
	fprintf(fpDat, "CPU stopped with residual norm %e in %f ms\n\n", res_norm_ref, t_h_ms);
    #endif

	/* solve Ax = b on single device */
	#if EXECUTE_SINGLE_GPU
    cublasGetError();
    int iter_single_gpu;
    unsigned int timer_d = 0;
    double t_d_ms = 0.0;    
    TYPE res_norm;
	start_timer(&timer_d);	
    if(!strcmp(solver_name, "CG"))
    {
        if(!strcmp(storage_name, "CSR"))       
            res_norm = cg_device<csr_format<TYPE>, TYPE>(fpDat, &spm_csr, h_b, h_x, max_iter, error_rel, &iter_single_gpu);
        else if(!strcmp(storage_name, "ELL"))
            res_norm = cg_device<ell_format<TYPE>, TYPE>(fpDat, &spm_ell, h_b, h_x, max_iter, error_rel, &iter_single_gpu);
    }
    else if(!strcmp(solver_name, "PCG"))
    {
        if(!strcmp(storage_name, "CSR")) 
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm = pcg_device<csr_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_csr, &Jacobi, h_b, h_x, max_iter, error_rel, &iter_single_gpu);
            else if(!strcmp(precon_name, "SSOR"))
                res_norm = pcg_device<csr_format<TYPE>, sub_ssor_precon<TYPE>, TYPE>(fpDat, &spm_csr, &sub_ssor, h_b, h_x, max_iter, error_rel, &iter_single_gpu);            
        }
        else if(!strcmp(storage_name, "ELL"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm = pcg_device<ell_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_ell, &Jacobi, h_b, h_x, max_iter, error_rel, &iter_single_gpu);        
            else if(!strcmp(precon_name, "SSOR"))
                res_norm = pcg_device<ell_format<TYPE>, sub_ssor_precon<TYPE>, TYPE>(fpDat, &spm_ell, &sub_ssor, h_b, h_x, max_iter, error_rel, &iter_single_gpu);        
        }
    }
    else if(!strcmp(solver_name, "PGMRES"))  
    {
        if(!strcmp(storage_name, "CSR")) 
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm = pgmres_device<csr_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_csr, &Jacobi, h_b, h_x, max_iter, error_rel, &iter_single_gpu);
            else if(!strcmp(precon_name, "SSOR"))
                res_norm = pgmres_device<csr_format<TYPE>, sub_ssor_precon<TYPE>, TYPE>(fpDat, &spm_csr, &sub_ssor, h_b, h_x, max_iter, error_rel, &iter_single_gpu);            
        }
        else if(!strcmp(storage_name, "ELL"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm = pgmres_device<ell_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_ell, &Jacobi, h_b, h_x, max_iter, error_rel, &iter_single_gpu);        
            else if(!strcmp(precon_name, "SSOR"))
                res_norm = pgmres_device<ell_format<TYPE>, sub_ssor_precon<TYPE>, TYPE>(fpDat, &spm_ell, &sub_ssor, h_b, h_x, max_iter, error_rel, &iter_single_gpu);        
        }        
    }
    stop_timer(timer_d, &t_d_ms);
	fprintf(fpDat, "Single GPU stopped with residual norm %e in %f ms\n\n", res_norm, t_d_ms);
    #if EXECUTE_CPU
    fprintf(fpDat, "Single GPU speedup: %f\n\n", t_h_ms / t_d_ms);
    #endif
    #endif

	/* solve Ax = b on multi-device */
	#if EXECUTE_MULTI_GPU
	cublasGetError();
    int iter_multi_gpu;
    unsigned int timer_md = 0;
	double t_md_ms = 0.0;
    TYPE res_norm_mul;
	start_timer(&timer_md);	
    if(!strcmp(solver_name, "CG"))
    {
        if(!strcmp(storage_name, "CSR"))  
            res_norm_mul = cg_multi_device<csr_format<TYPE>, TYPE>(fpDat, &spm_csr, h_b, mh_x, max_iter, error_rel, &iter_multi_gpu);
        else if(!strcmp(storage_name, "ELL"))
            res_norm_mul = cg_multi_device<ell_format<TYPE>, TYPE>(fpDat, &spm_ell, h_b, mh_x, max_iter, error_rel, &iter_multi_gpu);
    }
    else if(!strcmp(solver_name, "PCG"))
    {
        if(!strcmp(storage_name, "CSR"))  
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm_mul = pcg_multi_device<csr_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_csr, &Jacobi, h_b, mh_x, max_iter, error_rel, &iter_multi_gpu);
            else if(!strcmp(precon_name, "SSOR"))
                printf("Sorry, multi_device for SSOR preconditioner is not yet supported!\n");
                //res_norm_mul = pcg_multi_device<csr_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_csr, &spm_dia, h_b, mh_x, max_iter, error_rel, &iter_multi_gpu);            
        }
        else if(!strcmp(storage_name, "ELL"))
        {
            if(!strcmp(precon_name, "Jacobi"))
                res_norm_mul = pcg_multi_device<ell_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_ell, &Jacobi, h_b, mh_x, max_iter, error_rel, &iter_multi_gpu);
            else if(!strcmp(precon_name, "SSOR"))
                printf("Sorry, multi_device for SSOR preconditioner is not yet supported!\n");
                //res_norm_mul = pcg_multi_device<ell_format<TYPE>, Jacobi_precon<TYPE>, TYPE>(fpDat, &spm_ell, &spm_dia, h_b, mh_x, max_iter, error_rel, &iter_multi_gpu);            
        }
    }
	stop_timer(timer_md, &t_md_ms);
	fprintf(fpDat, "Multi-GPU stopped with residual norm %e in %f ms\n\n", res_norm_mul, t_md_ms);
    #if EXECUTE_CPU
    fprintf(fpDat, "Multi-GPU speedup(to CPU): %f\n", t_h_ms / t_md_ms);
    #endif
	#if EXECUTE_SINGLE_GPU
	fprintf(fpDat, "Multi-GPU speedup(to Single GPU): %f\n\n", t_d_ms / t_md_ms);
    #endif
    #endif

    /* compare the results */
    #if COMPARE_RESULT
    compare_result<TYPE>(fpDat, h_x_ref, h_x, ncol);
    #endif

    /* output */
    if (argc == 8) 
    { 
        strcpy(x_filename, argv[7]);
        write_x<TYPE>(h_x_ref, x_filename, ncol);
    }else
    {
        #if PRINT_X
        for (int i = 0; i < ncol; i++)
        printf("h_x_ref[%d]: %f\n", i, h_x_ref[i]);
        #endif
    }
    #if OUTPUT_TECPLOT
    output_tecplot<TYPE>(mh_x, x_num, y_num, hx, hy);
    #endif
    
	/* free host memory */
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaFreeHost(h_b));
    CUDA_SAFE_CALL(cudaFreeHost(h_x_ref));
    CUDA_SAFE_CALL(cudaFreeHost(h_x));
    CUDA_SAFE_CALL(cudaFreeHost(mh_x));    
    #else
    free(h_b);
	free(h_x_ref);
	free(h_x);
	free(mh_x);
    #endif
    if(!strcmp(storage_name, "CSR"))  
        clear_format<TYPE>(&spm_csr);
    else if(!strcmp(storage_name, "ELL"))
        clear_format<TYPE>(&spm_ell);

    if(!strcmp(precon_name, "Jacobi"))
        clear_precon<TYPE>(&Jacobi);
    else if(!strcmp(precon_name, "SSOR"))
        {
           clear_precon<TYPE>(&ssor);
           clear_precon<TYPE>(&sub_ssor);
        }        
    
    fclose(fpDat);	
    printf("All are OK!\n");
    return EXIT_SUCCESS;
}
