/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __SPM_DIAG_H__
#define __SPM_DIAG_H__

template <class T>
struct spm_diag
{
    int nrow;
    T *diag;
};

template <class T>
void get_diag(spmatrix<T> *spm, spm_diag<T> *spm_dia)
{
    int row, colomn;
    spm_dia->nrow = spm->nrow; 
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaMallocHost((void **)&spm_dia->diag, spm_dia->nrow * sizeof(T)));
    #else
    spm_dia->diag = (T *)malloc(sizeof(T) * spm_dia->nrow);
    #endif

    memset(spm_dia->diag, 0, spm_dia->nrow * sizeof(T));    
    for(int i = 0; i < spm->nnzero; i++) 
    {
        row = (spm->nzentry)[i].row_num;
        colomn = (spm->nzentry)[i].col_num;
        if(row == colomn)
        (spm_dia->diag)[row] = (spm->nzentry)[i].val;
    }
}

template <class T>
void clear_diag(spm_diag<T> *spm_dia) 
{
    #if USE_PAGE_LOCKED
    CUDA_SAFE_CALL(cudaFreeHost(spm_dia->diag));
    #else
    free(spm_dia->diag);
    #endif
}

#endif /* __SPM_DIAG_H__ */
