/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __SUB_SSOR_PRECON_H__
#define __SUB_SSOR_PRECON_H__

#include "blas.h"
#include "cutils.h"

template <class T>
struct sub_ssor_precon
{
    csc_tri_array<T> lower;
    T *diag;
    csc_tri_array<T> upper;
};

template <class T>
double get_precon_flop(sub_ssor_precon<T> *h_precon)
{
    int n = (h_precon->lower).nrow;
    int nnzero = (h_precon->lower).nnzero - n;
    double flop = 6.0 * n + 2.0 * nnzero;
    return flop;
}

template <class T>
void ssor_setup(csc_tri_array<T> *csc_lower, spm_diag<T> *spm_dia, csc_tri_array<T> *csc_upper, sub_ssor_precon<T> *ssor)
{
    (ssor->lower).ntri = csc_lower->ntri;
    (ssor->lower).csc_tri_entry = (csc_triangle<T> *)malloc(sizeof(csc_triangle<T>) * (ssor->lower).ntri);
    for(int i = 0; i < (ssor->lower).ntri; i++)
    {
        (ssor->lower).csc_tri_entry[i].nrow = (csc_lower->csc_tri_entry)[i].nrow;
        (ssor->lower).csc_tri_entry[i].ncol = (csc_lower->csc_tri_entry)[i].ncol;
        (ssor->lower).csc_tri_entry[i].nnzero = (csc_lower->csc_tri_entry)[i].nnzero;
        (ssor->lower).csc_tri_entry[i].nrow_per_col = (csc_lower->csc_tri_entry)[i].nrow_per_col;
        (ssor->lower).csc_tri_entry[i].col_ptr = (csc_lower->csc_tri_entry)[i].col_ptr;
        (ssor->lower).csc_tri_entry[i].row_ind = (csc_lower->csc_tri_entry)[i].row_ind;
        (ssor->lower).csc_tri_entry[i].nz_val = (csc_lower->csc_tri_entry)[i].nz_val;        
    }
    
    ssor->diag = spm_dia->diag;

    (ssor->upper).ntri = csc_upper->ntri;
    (ssor->upper).csc_tri_entry = (csc_triangle<T> *)malloc(sizeof(csc_triangle<T>) * (ssor->upper).ntri);
    for(int i = 0; i < (ssor->upper).ntri; i++)
    {
        (ssor->upper).csc_tri_entry[i].nrow = (csc_upper->csc_tri_entry)[i].nrow;
        (ssor->upper).csc_tri_entry[i].ncol = (csc_upper->csc_tri_entry)[i].ncol;
        (ssor->upper).csc_tri_entry[i].nnzero = (csc_upper->csc_tri_entry)[i].nnzero;
        (ssor->upper).csc_tri_entry[i].nrow_per_col = (csc_upper->csc_tri_entry)[i].nrow_per_col;
        (ssor->upper).csc_tri_entry[i].col_ptr = (csc_upper->csc_tri_entry)[i].col_ptr;
        (ssor->upper).csc_tri_entry[i].row_ind = (csc_upper->csc_tri_entry)[i].row_ind;
        (ssor->upper).csc_tri_entry[i].nz_val = (csc_upper->csc_tri_entry)[i].nz_val;        
    }    
}

template <class T>
void clear_precon(sub_ssor_precon<T> *ssor) 
{
    free((ssor->lower).csc_tri_entry);
    free(ssor->diag);
    free((ssor->upper).csc_tri_entry);
}



#endif /* __SUB_SSOR_PRECON_H__ */
