/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __TEXTURE_H__
#define __TEXTURE_H__

#include <cutil_inline.h>

texture<float, 1> tex_x_float;
// Use int2 to pull doubles through texture cache
texture<int2, 1>  tex_x_double;

void bind_vec(float *x) 
{   
   CUDA_SAFE_CALL(cudaBindTexture(NULL, tex_x_float, x));   
}

void bind_vec(double *x) 
{   
   CUDA_SAFE_CALL(cudaBindTexture(NULL, tex_x_double, x));   
} 

void unbind_vec(float *x) 
{ 
   CUDA_SAFE_CALL(cudaUnbindTexture(tex_x_float)); 
}

void unbind_vec(double *x) 
{   
   CUDA_SAFE_CALL(cudaUnbindTexture(tex_x_double)); 
}

__inline__ __device__ float fetch_vec(const int &i, float *x) 
{
   return tex1Dfetch(tex_x_float, i);
}

__inline__ __device__ double fetch_vec(const int &i, double *x)
{
    int2 v = tex1Dfetch(tex_x_double, i);
    return __hiloint2double(v.y, v.x);//avoid banks conflict
}

dim3 large_grid(const int num, const int dimBlock)
{
    const unsigned int dimGrid = (num + dimBlock - 1) / dimBlock;
    if(dimGrid <= 65535)
        return dim3(dimGrid);
    else 
    {
        //2D grid is required
        const unsigned int side = (unsigned int) ceil(sqrt((double)dimGrid));
        return dim3(side, side);
    }
}

#endif /* __TEXTURE_H__ */
