/*
 * CUDA for Sparse Solver
 * Author: Luo Li     li.luo@sub.siat.ac.cn
 */ 

#ifndef __TIMER_H__
#define __TIMER_H__

void start_timer(unsigned int *t)
{
    CUT_SAFE_CALL(cutCreateTimer(t));
    CUT_SAFE_CALL(cutStartTimer((*t)));
    return;
}

template <class T>
void stop_timer(unsigned int t, T *t_ms)
{
    CUT_SAFE_CALL(cutStopTimer(t));
    (*t_ms) = cutGetTimerValue(t);
    CUT_SAFE_CALL(cutDeleteTimer(t));
    return;
}

struct timer
{
    long long freq;
    long long clocks;
    cudaEvent_t start;
    cudaEvent_t stop;
};

void begin_timer(timer *timer) 
{
    timer->freq = 1000000;
    timer->clocks = 0;
    cudaEventCreate(&(timer->start));
    cudaEventRecord(timer->start, 0);
}

double end_timer(timer *timer) 
{
    float elaps;   
    cudaEventCreate(&(timer->stop));
    cudaEventRecord(timer->stop, 0);
    cudaEventSynchronize(timer->stop);
    cudaEventElapsedTime(&elaps, timer->start, timer->stop);
    cudaEventDestroy(timer->stop);
    timer->clocks = (long long)(elaps * 1000.0); 
    return (double)timer->clocks / (double)timer->freq; 
}

double elapsed_timer(timer *timer, double old_time) 
{
    end_timer(timer); 
    return ((double)timer->clocks / (double)timer->freq - old_time) ; 
}

#endif /* __TIMER_H__ */
